@@ -13,7 +13,7 @@ from buildbot.changes import base
 from buildbot.changes.changes import Change
 
 import xml.dom.minidom
-import urllib
+import os, urllib
 
 def _assert(condition, msg):
     if condition:
@@ -44,7 +44,7 @@ class SVNPoller(base.ChangeSource, util.ComparableMixin):
     compare_attrs = ["svnurl", "split_file_function",
                      "svnuser", "svnpasswd",
                      "pollinterval", "histmax",
-                     "svnbin", "category"]
+                     "svnbin", "category", "cachepath"]
 
     parent = None # filled in when we're added
     last_change = None
@@ -54,7 +54,8 @@ class SVNPoller(base.ChangeSource, util.ComparableMixin):
     def __init__(self, svnurl, split_file=None,
                  svnuser=None, svnpasswd=None,
                  pollinterval=10*60, histmax=100,
-                 svnbin='svn', revlinktmpl='', category=None):
+                 svnbin='svn', revlinktmpl='', category=None, 
+                 project=None, cachepath=None):
         """
         @type  svnurl: string
         @param svnurl: the SVN URL that describes the repository and
@@ -172,6 +173,16 @@ class SVNPoller(base.ChangeSource, util.ComparableMixin):
         @param category:     A single category associated with the changes that
                              could be used by schedulers watch for branches of a
                              certain name AND category.
+                             
+        @type  project       string
+        @param project       A single project that the changes are associated with
+                             the repository, added to the changes, for the use in 
+                             change filters
+
+        @type  cachepath     string
+        @param cachepath     A path to a file that can be used to store the last
+                             rev that was processed, so we can grab changes that
+                             happened while we were offline
         """
 
         if svnurl.endswith("/"):
@@ -183,6 +194,9 @@ class SVNPoller(base.ChangeSource, util.ComparableMixin):
 
         self.revlinktmpl = revlinktmpl
 
+        self.environ = os.environ.copy() # include environment variables
+                                         # required for ssh-agent auth
+
         self.svnbin = svnbin
         self.pollinterval = pollinterval
         self.histmax = histmax
@@ -190,6 +204,19 @@ class SVNPoller(base.ChangeSource, util.ComparableMixin):
         self.overrun_counter = 0
         self.loop = LoopingCall(self.checksvn)
         self.category = category
+        self.project = project
+
+        self.cachepath = cachepath
+        if self.cachepath and os.path.exists(self.cachepath):
+            try:
+                f = open(self.cachepath, "r")
+                self.last_change = int(f.read().strip())
+                log.msg("SVNPoller(%s) setting last_change to %s" % (self.svnurl, self.last_change))
+                f.close()
+            except:
+                self.cachepath = None
+                log.msg("SVNPoller(%s) cache file corrupt, skipping and not using" % self.svnurl)
+                log.err()
 
     def split_file(self, path):
         # use getattr() to avoid turning this function into a bound method,
@@ -252,7 +279,10 @@ class SVNPoller(base.ChangeSource, util.ComparableMixin):
             return defer.succeed(None)
         self.working = True
 
-        log.msg("SVNPoller polling")
+        if self.project:
+            log.msg("SVNPoller polling " + self.project)
+        else:
+            log.msg("SVNPoller polling")
         if not self._prefix:
             # this sets self._prefix when it finishes. It fires with
             # self._prefix as well, because that makes the unit tests easier
@@ -272,7 +302,7 @@ class SVNPoller(base.ChangeSource, util.ComparableMixin):
 
     def getProcessOutput(self, args):
         # this exists so we can override it during the unit tests
-        d = utils.getProcessOutput(self.svnbin, args, {})
+        d = utils.getProcessOutput(self.svnbin, args, self.environ)
         return d
 
     def get_root(self):
@@ -419,7 +449,12 @@ class SVNPoller(base.ChangeSource, util.ComparableMixin):
             #when     = time.mktime(time.strptime("%.19s" % when,
             #                                     "%Y-%m-%dT%H:%M:%S"))
             branches = {}
-            pathlist = el.getElementsByTagName("paths")[0]
+            try:
+                pathlist = el.getElementsByTagName("paths")[0]
+            except IndexError: # weird, we got an empty revision
+                log.msg("ignoring commit with no paths")
+                continue
+
             for p in pathlist.getElementsByTagName("path"):
                 action = p.getAttribute("action")
                 path = "".join([t.data for t in p.childNodes])
@@ -458,7 +493,8 @@ class SVNPoller(base.ChangeSource, util.ComparableMixin):
                                branch=branch,
                                revlink=revlink,
                                category=self.category,
-                               repository=self.svnurl)
+                               repository=self.svnurl,
+                               project = self.project)
                     changes.append(c)
 
         return changes
@@ -468,6 +504,11 @@ class SVNPoller(base.ChangeSource, util.ComparableMixin):
             self.parent.addChange(c)
 
     def finished_ok(self, res):
+        if self.cachepath:
+            f = open(self.cachepath, "w")
+            f.write(str(self.last_change))
+            f.close()
+
         log.msg("SVNPoller finished polling %s" % res)
         assert self.working
         self.working = False
