@@ -1,13 +1,9 @@
 # -*- test-case-name: buildbot.test.test_run -*-
 
 import os
-signal = None
-try:
-    import signal
-except ImportError:
-    pass
-from cPickle import load
-import warnings
+import signal
+import time
+import textwrap
 
 from zope.interface import implements
 from twisted.python import log, components
@@ -16,21 +12,25 @@ from twisted.internet import defer, reactor
 from twisted.spread import pb
 from twisted.cred import portal, checkers
 from twisted.application import service, strports
-from twisted.persisted import styles
+from twisted.application.internet import TimerService
 
 import buildbot
 # sibling imports
-from buildbot.util import now, safeTranslate
+from buildbot.util import now, safeTranslate, eventual
 from buildbot.pbutil import NewCredPerspective
 from buildbot.process.builder import Builder, IDLE
-from buildbot.process.base import BuildRequest
-from buildbot.status.builder import Status
-from buildbot.changes.changes import Change, ChangeMaster, TestChangeMaster
-from buildbot.sourcestamp import SourceStamp
-from buildbot.buildslave import BuildSlave
+from buildbot.status.builder import Status, BuildSetStatus
+from buildbot.changes.changes import Change
+from buildbot.changes.manager import ChangeManager
 from buildbot import interfaces, locks
 from buildbot.process.properties import Properties
 from buildbot.config import BuilderConfig
+from buildbot.process.builder import BuilderControl
+from buildbot.db.dbspec import DBSpec
+from buildbot.db import connector, exceptions
+from buildbot.db.schema.manager import DBSchemaManager
+from buildbot.schedulers.manager import SchedulerManager
+from buildbot.util.loop import DelegateLoop
 
 ########################################
 
@@ -42,6 +42,7 @@ class BotMaster(service.MultiService):
     """
 
     debug = 0
+    reactor = reactor
 
     def __init__(self):
         service.MultiService.__init__(self)
@@ -72,6 +73,97 @@ class BotMaster(service.MultiService):
         # traversal
         self.prioritizeBuilders = None
 
+        self.loop = DelegateLoop(self._get_processors)
+        self.loop.setServiceParent(self)
+
+        self.shuttingDown = False
+
+    def setMasterName(self, name, incarnation):
+        self.master_name = name
+        self.master_incarnation = incarnation
+
+    def cleanShutdown(self):
+        if self.shuttingDown:
+            return
+        log.msg("Initiating clean shutdown")
+        self.shuttingDown = True
+
+        # Wait for all builds to finish
+        l = []
+        for builder in self.builders.values():
+            for build in builder.builder_status.getCurrentBuilds():
+                l.append(build.waitUntilFinished())
+        if len(l) == 0:
+            log.msg("No running jobs, starting shutdown immediately")
+            self.loop.trigger()
+            d = self.loop.when_quiet()
+        else:
+            log.msg("Waiting for %i build(s) to finish" % len(l))
+            d = defer.DeferredList(l)
+            d.addCallback(lambda ign: self.loop.when_quiet())
+
+        # Flush the eventual queue
+        d.addCallback(eventual.flushEventualQueue)
+
+        # Finally, shut the whole process down
+        def shutdown(ign):
+            # Double check that we're still supposed to be shutting down
+            # The shutdown may have been cancelled!
+            if self.shuttingDown:
+                # Check that there really aren't any running builds
+                for builder in self.builders.values():
+                    n = len(builder.builder_status.getCurrentBuilds())
+                    if n > 0:
+                        log.msg("Not shutting down, builder %s has %i builds running" % (builder, n))
+                        log.msg("Trying shutdown sequence again")
+                        self.shuttingDown = False
+                        self.cleanShutdown()
+                        return
+                log.msg("Stopping reactor")
+                self.reactor.stop()
+        d.addCallback(shutdown)
+        return d
+
+    def cancelCleanShutdown(self):
+        if not self.shuttingDown:
+            return
+        log.msg("Cancelling clean shutdown")
+        self.shuttingDown = False
+
+    def _sortfunc(self, b1, b2):
+        t1 = b1.getOldestRequestTime()
+        t2 = b2.getOldestRequestTime()
+        # If t1 or t2 is None, then there are no build requests,
+        # so sort it at the end
+        if t1 is None:
+            return 1
+        if t2 is None:
+            return -1
+        return cmp(t1, t2)
+
+    def _sort_builders(self, parent, builders):
+        return sorted(builders, self._sortfunc)
+
+    def _get_processors(self):
+        if self.shuttingDown:
+            return []
+        builders = self.builders.values()
+        sorter = self.prioritizeBuilders or self._sort_builders
+        try:
+            builders = sorter(self.parent, builders)
+        except:
+            log.msg("Exception prioritizing builders")
+            log.err(Failure())
+            # leave them in the original order
+        return [b.run for b in builders]
+
+    def trigger_add_buildrequest(self, category, *brids):
+        # a buildrequest has been added or resubmitted
+        self.loop.trigger()
+    def triggerNewBuildCheck(self):
+        # called when a build finishes, or a slave attaches
+        self.loop.trigger()
+
     # these four are convenience functions for testing
 
     def waitUntilBuilderAttached(self, name):
@@ -181,51 +273,37 @@ class BotMaster(service.MultiService):
         return allBuilders
 
     def setBuilders(self, builders):
+        # TODO: remove self.builders and just use the Service hierarchy to
+        # keep track of active builders. We could keep self.builderNames to
+        # retain ordering, if it seems important.
         self.builders = {}
         self.builderNames = []
-        for b in builders:
-            for slavename in b.slavenames:
-                # this is actually validated earlier
-                assert slavename in self.slaves
-            self.builders[b.name] = b
-            self.builderNames.append(b.name)
-            b.setBotmaster(self)
-        d = self._updateAllSlaves()
+        d = defer.DeferredList([b.disownServiceParent() for b in list(self)
+                                if isinstance(b, Builder)],
+                               fireOnOneErrback=True)
+        def _add(ign):
+            log.msg("setBuilders._add: %s %s" % (list(self), builders))
+            for b in builders:
+                for slavename in b.slavenames:
+                    # this is actually validated earlier
+                    assert slavename in self.slaves
+                self.builders[b.name] = b
+                self.builderNames.append(b.name)
+                b.setBotmaster(self)
+                b.setServiceParent(self)
+        d.addCallback(_add)
+        d.addCallback(lambda ign: self._updateAllSlaves())
         return d
 
     def _updateAllSlaves(self):
         """Notify all buildslaves about changes in their Builders."""
-        dl = [s.updateSlave() for s in self.slaves.values()]
+        dl = []
+        for s in self.slaves.values():
+            d = s.updateSlave()
+            d.addErrback(log.err)
+            dl.append(d)
         return defer.DeferredList(dl)
 
-    def maybeStartAllBuilds(self):
-        builders = self.builders.values()
-        if self.prioritizeBuilders is not None:
-            try:
-                builders = self.prioritizeBuilders(self.parent, builders)
-            except:
-                log.msg("Exception prioritizing builders")
-                log.err(Failure())
-                return
-        else:
-            def _sortfunc(b1, b2):
-                t1 = b1.getOldestRequestTime()
-                t2 = b2.getOldestRequestTime()
-                # If t1 or t2 is None, then there are no build requests,
-                # so sort it at the end
-                if t1 is None:
-                    return 1
-                if t2 is None:
-                    return -1
-                return cmp(t1, t2)
-            builders.sort(_sortfunc)
-        try:
-            for b in builders:
-                b.maybeStartBuild()
-        except:
-            log.msg("Exception starting builds")
-            log.err(Failure())
-
     def shouldMergeRequests(self, builder, req1, req2):
         """Determine whether two BuildRequests should be merged for
         the given builder.
@@ -235,8 +313,51 @@ class BotMaster(service.MultiService):
             return self.mergeRequests(builder, req1, req2)
         return req1.canBeMergedWith(req2)
 
-    def getPerspective(self, slavename):
-        return self.slaves[slavename]
+    def getPerspective(self, mind, slavename):
+        sl = self.slaves[slavename]
+        if not sl:
+            return None
+
+        # record when this connection attempt occurred
+        sl.recordConnectTime()
+
+        if sl.isConnected():
+            # uh-oh, we've got a duplicate slave. The most likely
+            # explanation is that the slave is behind a slow link, thinks we
+            # went away, and has attempted to reconnect, so we've got two
+            # "connections" from the same slave.  The old may not be stale at this
+            # point, if there are two slave proceses out there with the same name,
+            # so instead of booting the old (which may be in the middle of a build),
+            # we reject the new connection and ping the old slave.
+            log.msg("duplicate slave %s; rejecting new slave and pinging old" % sl.slavename)
+
+            # just in case we've got two identically-configured slaves,
+            # report the IP addresses of both so someone can resolve the
+            # squabble
+            old_tport = sl.slave.broker.transport
+            new_tport = mind.broker.transport
+            log.msg("old slave was connected from", old_tport.getPeer())
+            log.msg("new slave is from", new_tport.getPeer())
+
+            # ping the old slave.  If this kills it, then the new slave will connect
+            # again and everyone will be happy.
+            d = sl.slave.callRemote("print", "master got a duplicate connection; keeping this one")
+
+            # now return a dummy avatar and kill the new connection in 5
+            # seconds, thereby giving the ping a bit of time to kill the old
+            # connection, if necessary
+            def kill():
+                log.msg("killing new slave on", new_tport.getPeer())
+                new_tport.loseConnection()
+            reactor.callLater(5, kill)
+            class DummyAvatar(pb.Avatar):
+                def attached(self, *args):
+                    pass
+                def detached(self, *args):
+                    pass
+            return DummyAvatar()
+
+        return sl
 
     def shutdownSlaves(self):
         # TODO: make this into a bot method rather than a builder method
@@ -247,7 +368,7 @@ class BotMaster(service.MultiService):
         for b in self.builders.values():
             b.builder_status.addPointEvent(["master", "shutdown"])
             b.builder_status.saveYourself()
-        return service.Service.stopService(self)
+        return service.MultiService.stopService(self)
 
     def getLockByID(self, lockid):
         """Convert a Lock identifier into an actual Lock instance.
@@ -274,13 +395,13 @@ class DebugPerspective(NewCredPerspective):
         pass
 
     def perspective_requestBuild(self, buildername, reason, branch, revision, properties={}):
+        from buildbot.sourcestamp import SourceStamp
         c = interfaces.IControl(self.master)
         bc = c.getBuilder(buildername)
         ss = SourceStamp(branch, revision)
         bpr = Properties()
         bpr.update(properties, "remote requestBuild")
-        br = BuildRequest(reason, ss, builderName=buildername, properties=bpr)
-        bc.requestBuild(br)
+        bc.submitBuildRequest(ss, reason, bpr)
 
     def perspective_pingBuilder(self, buildername):
         c = interfaces.IControl(self.master)
@@ -288,9 +409,11 @@ class DebugPerspective(NewCredPerspective):
         bc.ping()
 
     def perspective_fakeChange(self, file, revision=None, who="fakeUser",
-                               branch=None):
+                               branch=None, repository="", 
+                               project=""):
         change = Change(who, [file], "some fake comments\n",
-                        branch=branch, revision=revision)
+                        branch=branch, revision=revision,
+                        repository=repository, project=project)
         c = interfaces.IControl(self.master)
         c.addChange(change)
 
@@ -337,6 +460,8 @@ class Dispatcher:
         afactory = self.names.get(avatarID)
         if afactory:
             p = afactory.getPerspective()
+        elif avatarID == "change":
+            raise ValueError("no PBChangeSource installed")
         elif avatarID == "debug":
             p = DebugPerspective()
             p.master = self.master
@@ -346,31 +471,27 @@ class Dispatcher:
         else:
             # it must be one of the buildslaves: no other names will make it
             # past the checker
-            p = self.botmaster.getPerspective(avatarID)
+            p = self.botmaster.getPerspective(mind, avatarID)
 
         if not p:
             raise ValueError("no perspective for '%s'" % avatarID)
 
         d = defer.maybeDeferred(p.attached, mind)
-        d.addCallback(self._avatarAttached, mind)
+        def _avatarAttached(_, mind):
+            return (pb.IPerspective, p, lambda: p.detached(mind))
+        d.addCallback(_avatarAttached, mind)
         return d
 
-    def _avatarAttached(self, p, mind):
-        return (pb.IPerspective, p, lambda p=p,mind=mind: p.detached(mind))
 
 ########################################
 
-# service hierarchy:
-#  BuildMaster
-#   BotMaster
-#   ChangeMaster
-#    all IChangeSource objects
-#   StatusClientService
-#   TCPClient(self.ircFactory)
-#   TCPServer(self.slaveFactory) -> dispatcher.requestAvatar
-#   TCPServer(self.site)
-#   UNIXServer(ResourcePublisher(self.site))
+class _Unset: pass  # marker
 
+class LogRotation: 
+    '''holds log rotation parameters (for WebStatus)'''
+    def __init__(self):
+        self.rotateLength = 1 * 1000 * 1000 
+        self.maxRotatedFiles = 10
 
 class BuildMaster(service.MultiService):
     debug = 0
@@ -382,7 +503,7 @@ class BuildMaster(service.MultiService):
     change_svc = None
     properties = Properties()
 
-    def __init__(self, basedir, configFileName="master.cfg"):
+    def __init__(self, basedir, configFileName="master.cfg", db_spec=None):
         service.MultiService.__init__(self)
         self.setName("buildmaster")
         self.basedir = basedir
@@ -404,25 +525,39 @@ class BuildMaster(service.MultiService):
         self.slavePortnum = None
         self.slavePort = None
 
+        self.change_svc = ChangeManager()
+        self.change_svc.setServiceParent(self)
+        self.dispatcher.changemaster = self.change_svc
+
+        try:
+            hostname = os.uname()[1] # only on unix
+        except AttributeError:
+            hostname = "?"
+        self.master_name = "%s:%s" % (hostname, os.path.abspath(self.basedir))
+        self.master_incarnation = "pid%d-boot%d" % (os.getpid(), time.time())
+
         self.botmaster = BotMaster()
         self.botmaster.setName("botmaster")
+        self.botmaster.setMasterName(self.master_name, self.master_incarnation)
         self.botmaster.setServiceParent(self)
-        dispatcher.botmaster = self.botmaster
+        self.dispatcher.botmaster = self.botmaster
 
         self.status = Status(self.botmaster, self.basedir)
-
         self.statusTargets = []
 
-        # this ChangeMaster is a dummy, only used by tests. In the real
-        # buildmaster, where the BuildMaster instance is activated
-        # (startService is called) by twistd, this attribute is overwritten.
-        self.useChanges(TestChangeMaster())
+        self.db = None
+        self.db_url = None
+        self.db_poll_interval = _Unset
+        if db_spec:
+            self.loadDatabase(db_spec)
 
         self.readConfig = False
+        
+        # create log_rotation object and set default parameters (used by WebStatus)
+        self.log_rotation = LogRotation()
 
     def startService(self):
         service.MultiService.startService(self)
-        self.loadChanges() # must be done before loading the config file
         if not self.readConfig:
             # TODO: consider catching exceptions during this call to
             # loadTheConfigFile and bailing (reactor.stop) if it fails,
@@ -430,35 +565,12 @@ class BuildMaster(service.MultiService):
             # the config file, and it would be nice for the user to discover
             # this quickly.
             self.loadTheConfigFile()
-        if signal and hasattr(signal, "SIGHUP"):
+        if hasattr(signal, "SIGHUP"):
             signal.signal(signal.SIGHUP, self._handleSIGHUP)
         for b in self.botmaster.builders.values():
             b.builder_status.addPointEvent(["master", "started"])
             b.builder_status.saveYourself()
 
-    def useChanges(self, changes):
-        if self.change_svc:
-            # TODO: can return a Deferred
-            self.change_svc.disownServiceParent()
-        self.change_svc = changes
-        self.change_svc.basedir = self.basedir
-        self.change_svc.setName("changemaster")
-        self.dispatcher.changemaster = self.change_svc
-        self.change_svc.setServiceParent(self)
-
-    def loadChanges(self):
-        filename = os.path.join(self.basedir, "changes.pck")
-        try:
-            changes = load(open(filename, "rb"))
-            styles.doUpgrade()
-        except IOError:
-            log.msg("changes.pck missing, using new one")
-            changes = ChangeMaster()
-        except EOFError:
-            log.msg("corrupted changes.pck, using new one")
-            changes = ChangeMaster()
-        self.useChanges(changes)
-
     def _handleSIGHUP(self, *args):
         reactor.callLater(0, self.loadTheConfigFile)
 
@@ -485,21 +597,27 @@ class BuildMaster(service.MultiService):
             return
 
         try:
-            self.loadConfig(f)
+            d = self.loadConfig(f)
         except:
             log.msg("error during loadConfig")
             log.err()
             log.msg("The new config file is unusable, so I'll ignore it.")
             log.msg("I will keep using the previous config file instead.")
+            return # sorry unit tests
         f.close()
+        return d # for unit tests
 
-    def loadConfig(self, f):
+    def loadConfig(self, f, check_synchronously_only=False):
         """Internal function to load a specific configuration file. Any
         errors in the file will be signalled by raising an exception.
 
-        @return: a Deferred that will fire (with None) when the configuration
-        changes have been completed. This may involve a round-trip to each
-        buildslave that was involved."""
+        If check_synchronously_only=True, I will return (with None)
+        synchronously, after checking the config file for sanity, or raise an
+        exception. I may also emit some DeprecationWarnings.
+
+        If check_synchronously_only=False, I will return a Deferred that
+        fires (with None) when the configuration changes have been completed.
+        This may involve a round-trip to each buildslave that was involved."""
 
         localDict = {'basedir': os.path.expanduser(self.basedir)}
         try:
@@ -515,15 +633,15 @@ class BuildMaster(service.MultiService):
             log.err("config file must define BuildmasterConfig")
             raise
 
-        known_keys = ("bots", "slaves",
-                      "sources", "change_source",
+        known_keys = ("slaves", "change_source",
                       "schedulers", "builders", "mergeRequests",
                       "slavePortnum", "debugPassword", "logCompressionLimit",
                       "manhole", "status", "projectName", "projectURL",
                       "buildbotURL", "properties", "prioritizeBuilders",
-                      "eventHorizon", "buildCacheSize", "logHorizon", "buildHorizon",
-                      "changeHorizon", "logMaxSize", "logMaxTailSize",
-                      "logCompressionMethod",
+                      "eventHorizon", "buildCacheSize", "changeCacheSize",
+                      "logHorizon", "buildHorizon", "changeHorizon",
+                      "logMaxSize", "logMaxTailSize", "logCompressionMethod",
+                      "db_url", "multiMaster", "db_poll_interval",
                       )
         for k in config.keys():
             if k not in known_keys:
@@ -538,6 +656,8 @@ class BuildMaster(service.MultiService):
             #change_source = config['change_source']
 
             # optional
+            db_url = config.get("db_url", "sqlite:///state.sqlite")
+            db_poll_interval = config.get("db_poll_interval", None)
             debugPassword = config.get('debugPassword')
             manhole = config.get('manhole')
             status = config.get('status', [])
@@ -546,7 +666,8 @@ class BuildMaster(service.MultiService):
             buildbotURL = config.get('buildbotURL')
             properties = config.get('properties', {})
             buildCacheSize = config.get('buildCacheSize', None)
-            eventHorizon = config.get('eventHorizon', None)
+            changeCacheSize = config.get('changeCacheSize', None)
+            eventHorizon = config.get('eventHorizon', 50)
             logHorizon = config.get('logHorizon', None)
             buildHorizon = config.get('buildHorizon', None)
             logCompressionLimit = config.get('logCompressionLimit', 4*1024)
@@ -574,30 +695,28 @@ class BuildMaster(service.MultiService):
             if changeHorizon is not None and not isinstance(changeHorizon, int):
                 raise ValueError("changeHorizon needs to be an int")
 
-        except KeyError, e:
+            multiMaster = config.get("multiMaster", False)
+
+        except KeyError:
             log.msg("config dictionary is missing a required parameter")
             log.msg("leaving old configuration in place")
             raise
 
-        #if "bots" in config:
-        #    raise KeyError("c['bots'] is no longer accepted")
+        if "sources" in config:
+            m = ("c['sources'] is deprecated as of 0.7.6 and is no longer "
+                 "accepted in >= 0.8.0 . Please use c['change_source'] instead.")
+            raise KeyError(m)
+
+        if "bots" in config:
+            m = ("c['bots'] is deprecated as of 0.7.6 and is no longer "
+                 "accepted in >= 0.8.0 . Please use c['slaves'] instead.")
+            raise KeyError(m)
 
         slaves = config.get('slaves', [])
-        if "bots" in config:
-            m = ("c['bots'] is deprecated as of 0.7.6 and will be "
-                 "removed by 0.8.0 . Please use c['slaves'] instead.")
-            log.msg(m)
-            warnings.warn(m, DeprecationWarning)
-            for name, passwd in config['bots']:
-                slaves.append(BuildSlave(name, passwd))
-
-        if "bots" not in config and "slaves" not in config:
-            log.msg("config dictionary must have either 'bots' or 'slaves'")
+        if "slaves" not in config:
+            log.msg("config dictionary must have a 'slaves' key")
             log.msg("leaving old configuration in place")
-            raise KeyError("must have either 'bots' or 'slaves'")
-
-        #if "sources" in config:
-        #    raise KeyError("c['sources'] is no longer accepted")
+            raise KeyError("must have a 'slaves' key")
 
         if changeHorizon is not None:
             self.change_svc.changeHorizon = changeHorizon
@@ -607,13 +726,6 @@ class BuildMaster(service.MultiService):
             change_sources = change_source
         else:
             change_sources = [change_source]
-        if "sources" in config:
-            m = ("c['sources'] is deprecated as of 0.7.6 and will be "
-                 "removed by 0.8.0 . Please use c['change_source'] instead.")
-            log.msg(m)
-            warnings.warn(m, DeprecationWarning)
-            for s in config['sources']:
-                change_sources.append(s)
 
         # do some validation first
         for s in slaves:
@@ -623,6 +735,12 @@ class BuildMaster(service.MultiService):
                     "reserved name '%s' used for a bot" % s.slavename)
         if config.has_key('interlocks'):
             raise KeyError("c['interlocks'] is no longer accepted")
+        assert self.db_url is None or db_url == self.db_url, \
+                "Cannot change db_url after master has started"
+        assert db_poll_interval is None or isinstance(db_poll_interval, int), \
+               "db_poll_interval must be an integer: seconds between polls"
+        assert self.db_poll_interval is _Unset or db_poll_interval == self.db_poll_interval, \
+               "Cannot change db_poll_interval after master has started"
 
         assert isinstance(change_sources, (list, tuple))
         for s in change_sources:
@@ -672,11 +790,13 @@ class BuildMaster(service.MultiService):
                 log.err(errmsg)
                 raise ValueError(errmsg)
 
-            # Fix the dictionnary with default values, in case this wasn't
+            # Fix the dictionary with default values, in case this wasn't
             # specified with a BuilderConfig object (which sets the same defaults)
             b.setdefault('builddir', safeTranslate(b['name']))
             b.setdefault('slavebuilddir', b['builddir'])
-
+            b.setdefault('buildHorizon', buildHorizon)
+            b.setdefault('logHorizon', logHorizon)
+            b.setdefault('eventHorizon', eventHorizon)
             if b['builddir'] in dirnames:
                 raise ValueError("builder %s reuses builddir %s"
                                  % (b['name'], b['builddir']))
@@ -686,22 +806,21 @@ class BuildMaster(service.MultiService):
         schedulernames = []
         for s in schedulers:
             for b in s.listBuilderNames():
-                assert b in buildernames, \
-                       "%s uses unknown builder %s" % (s, b)
+                # Skip checks for builders in multimaster mode
+                if not multiMaster:
+                    assert b in buildernames, \
+                           "%s uses unknown builder %s" % (s, b)
                 if b in unscheduled_buildernames:
                     unscheduled_buildernames.remove(b)
 
             if s.name in schedulernames:
-                # TODO: schedulers share a namespace with other Service
-                # children of the BuildMaster node, like status plugins, the
-                # Manhole, the ChangeMaster, and the BotMaster (although most
-                # of these don't have names)
                 msg = ("Schedulers must have unique names, but "
                        "'%s' was a duplicate" % (s.name,))
                 raise ValueError(msg)
             schedulernames.append(s.name)
 
-        if unscheduled_buildernames:
+        # Skip the checks for builders in multimaster mode
+        if not multiMaster and unscheduled_buildernames:
             log.msg("Warning: some Builders have no Schedulers to drive them:"
                     " %s" % (unscheduled_buildernames,))
 
@@ -741,6 +860,8 @@ class BuildMaster(service.MultiService):
         if type(slavePortnum) is int:
             slavePortnum = "tcp:%d" % slavePortnum
 
+        if check_synchronously_only:
+            return
         # now we're committed to implementing the new configuration, so do
         # it atomically
         # TODO: actually, this is spread across a couple of Deferreds, so it
@@ -774,10 +895,15 @@ class BuildMaster(service.MultiService):
             self.botmaster.prioritizeBuilders = prioritizeBuilders
 
         self.buildCacheSize = buildCacheSize
+        self.changeCacheSize = changeCacheSize
         self.eventHorizon = eventHorizon
         self.logHorizon = logHorizon
         self.buildHorizon = buildHorizon
 
+        # Set up the database
+        d.addCallback(lambda res:
+                      self.loadConfig_Database(db_url, db_poll_interval))
+
         # self.slaves: Disconnect any that were attached and removed from the
         # list. Update self.checker with the new list of passwords, including
         # debug/change/status.
@@ -811,7 +937,8 @@ class BuildMaster(service.MultiService):
         d.addCallback(lambda res: self.loadConfig_status(status))
 
         # Schedulers are added after Builders in case they start right away
-        d.addCallback(lambda res: self.loadConfig_Schedulers(schedulers))
+        d.addCallback(lambda res:
+                      self.scheduler_manager.updateSchedulers(schedulers))
         # and Sources go after Schedulers for the same reason
         d.addCallback(lambda res: self.loadConfig_Sources(change_sources))
 
@@ -837,11 +964,67 @@ class BuildMaster(service.MultiService):
             self.readConfig = True
             log.msg("configuration update complete")
         d.addCallback(_done)
-        d.addCallback(lambda res: self.botmaster.maybeStartAllBuilds())
+        d.addCallback(lambda res: self.botmaster.triggerNewBuildCheck())
+        d.addErrback(log.err)
         return d
 
+    def loadDatabase(self, db_spec, db_poll_interval=None):
+        if self.db:
+            return
+
+        # make sure it's up to date
+        sm = DBSchemaManager(db_spec, self.basedir)
+        if not sm.is_current():
+            raise exceptions.DatabaseNotReadyError, textwrap.dedent("""
+                The Buildmaster database needs to be upgraded before this version of buildbot
+                can run.  Use the following command-line
+                    buildbot upgrade-master path/to/master
+                to upgrade the database, and try starting the buildmaster again.  You may want
+                to make a backup of your buildmaster before doing so.  If you are using MySQL,
+                you must specify the connector string on the upgrade-master command line:
+                    buildbot upgrade-master --db=<db-connector-string> path/to/master
+                """)
+
+        self.db = connector.DBConnector(db_spec)
+        if self.changeCacheSize:
+            self.db.setChangeCacheSize(self.changeCacheSize)
+        self.db.start()
+
+        self.botmaster.db = self.db
+        self.status.setDB(self.db)
+
+        self.db.subscribe_to("add-buildrequest",
+                             self.botmaster.trigger_add_buildrequest)
+
+        sm = SchedulerManager(self, self.db, self.change_svc)
+        self.db.subscribe_to("add-change", sm.trigger_add_change)
+        self.db.subscribe_to("modify-buildset", sm.trigger_modify_buildset)
+
+        self.scheduler_manager = sm
+        sm.setServiceParent(self)
+
+        # Set db_poll_interval (perhaps to 30 seconds) if you are using
+        # multiple buildmasters that share a common database, such that the
+        # masters need to discover what each other is doing by polling the
+        # database. TODO: this will be replaced by the DBNotificationServer.
+        if db_poll_interval:
+            # it'd be nice if TimerService let us set now=False
+            t1 = TimerService(db_poll_interval, sm.trigger)
+            t1.setServiceParent(self)
+            t2 = TimerService(db_poll_interval, self.botmaster.loop.trigger)
+            t2.setServiceParent(self)
+        # adding schedulers (like when loadConfig happens) will trigger the
+        # scheduler loop at least once, which we need to jump-start things
+        # like Periodic.
+
+    def loadConfig_Database(self, db_url, db_poll_interval):
+        self.db_url = db_url
+        self.db_poll_interval = db_poll_interval
+        db_spec = DBSpec.from_url(db_url, self.basedir)
+        self.loadDatabase(db_spec, db_poll_interval)
+
     def loadConfig_Slaves(self, new_slaves):
-        # set up the Checker with the names and passwords of all valid bots
+        # set up the Checker with the names and passwords of all valid slaves
         self.checker.users = {} # violates abstraction, oh well
         for s in new_slaves:
             self.checker.addUser(s.slavename, s.password)
@@ -865,32 +1048,7 @@ class BuildMaster(service.MultiService):
         return d
 
     def allSchedulers(self):
-        return [child for child in self
-                if interfaces.IScheduler.providedBy(child)]
-
-
-    def loadConfig_Schedulers(self, newschedulers):
-        oldschedulers = self.allSchedulers()
-        removed = [s for s in oldschedulers if s not in newschedulers]
-        added = [s for s in newschedulers if s not in oldschedulers]
-        dl = [defer.maybeDeferred(s.disownServiceParent) for s in removed]
-        def addNewOnes(res):
-            log.msg("adding %d new schedulers, removed %d" % 
-                    (len(added), len(dl)))
-            for s in added:
-                s.setServiceParent(self)
-        d = defer.DeferredList(dl, fireOnOneErrback=1)
-        d.addCallback(addNewOnes)
-        if removed or added:
-            # notify Downstream schedulers to potentially pick up
-            # new schedulers now that we have removed and added some
-            def updateDownstreams(res):
-                log.msg("notifying downstream schedulers of changes")
-                for s in newschedulers:
-                    if interfaces.IDownstreamScheduler.providedBy(s):
-                        s.checkUpstreamScheduler()
-            d.addCallback(updateDownstreams)
-        return d
+        return list(self.scheduler_manager)
 
     def loadConfig_Builders(self, newBuilderData):
         somethingChanged = False
@@ -985,25 +1143,30 @@ class BuildMaster(service.MultiService):
 
 
     def addChange(self, change):
-        for s in self.allSchedulers():
-            s.addChange(change)
+        self.db.addChangeToDatabase(change)
+        self.status.changeAdded(change)
 
-    def submitBuildSet(self, bs):
+    def triggerSlaveManager(self):
+        self.botmaster.triggerNewBuildCheck()
+
+    def submitBuildSet(self, builderNames, ss, reason, props=None, now=False):
         # determine the set of Builders to use
-        builders = []
-        for name in bs.builderNames:
+        for name in builderNames:
             b = self.botmaster.builders.get(name)
-            if b:
-                if b not in builders:
-                    builders.append(b)
-                continue
-            # TODO: add aliases like 'all'
-            raise KeyError("no such builder named '%s'" % name)
+            if not b:
+                raise KeyError("no such builder named '%s'" % name)
+            if now and not b.slaves:
+                raise interfaces.NoSlaveError
+        if props is None:
+            props = Properties()
+        bsid = self.db.runInteractionNow(self._txn_submitBuildSet,
+                                         builderNames, ss, reason, props)
+        return BuildSetStatus(bsid, self.status, self.db)
 
-        # now tell the BuildSet to create BuildRequests for all those
-        # Builders and submit them
-        bs.start(builders)
-        self.status.buildsetSubmitted(bs.status)
+    def _txn_submitBuildSet(self, t, builderNames, ss, reason, props):
+        ssid = self.db.get_sourcestampid(ss, t)
+        bsid = self.db.create_buildset(ssid, reason, props, builderNames, t)
+        return bsid
 
 
 class Control:
@@ -1013,16 +1176,13 @@ class Control:
         self.master = master
 
     def addChange(self, change):
-        self.master.change_svc.addChange(change)
+        self.master.addChange(change)
 
-    def submitBuildSet(self, bs):
-        self.master.submitBuildSet(bs)
+    def submitBuildSet(self, builderNames, ss, reason, props=None, now=False):
+        return self.master.submitBuildSet(builderNames, ss, reason, props, now)
 
     def getBuilder(self, name):
         b = self.master.botmaster.builders[name]
-        return interfaces.IBuilderControl(b)
+        return BuilderControl(b, self)
 
 components.registerAdapter(Control, BuildMaster, interfaces.IControl)
-
-# so anybody who can get a handle on the BuildMaster can cause a build with:
-#  IControl(master).getBuilder("full-2.3").requestBuild(buildrequest)
