@@ -1,6 +1,7 @@
 from buildbot.process.buildstep import LoggingBuildStep, SUCCESS, FAILURE, EXCEPTION
 from buildbot.process.properties import Properties
-from buildbot.scheduler import Triggerable
+from buildbot.schedulers.triggerable import Triggerable
+from twisted.python import log
 from twisted.internet import defer
 
 class Trigger(LoggingBuildStep):
@@ -89,13 +90,11 @@ class Trigger(LoggingBuildStep):
         triggered_schedulers = []
 
         # TODO: don't fire any schedulers if we discover an unknown one
-        dl = []
         for scheduler in self.schedulerNames:
             scheduler = properties.render(scheduler)
             if all_schedulers.has_key(scheduler):
                 sch = all_schedulers[scheduler]
                 if isinstance(sch, Triggerable):
-                    dl.append(sch.trigger(ss, set_props=props_to_set))
                     triggered_schedulers.append(scheduler)
                 else:
                     unknown_schedulers.append(scheduler)
@@ -104,10 +103,13 @@ class Trigger(LoggingBuildStep):
 
         if unknown_schedulers:
             self.step_status.setText(['no scheduler:'] + unknown_schedulers)
-            rc = FAILURE
-        else:
-            rc = SUCCESS
-            self.step_status.setText(['triggered'] + triggered_schedulers)
+            return self.finished(FAILURE)
+
+        dl = []
+        for scheduler in triggered_schedulers:
+            sch = all_schedulers[scheduler]
+            dl.append(sch.trigger(ss, set_props=props_to_set))
+        self.step_status.setText(['triggered'] + triggered_schedulers)
 
         if self.waitForFinish:
             d = defer.DeferredList(dl, consumeErrors=1)
@@ -116,12 +118,13 @@ class Trigger(LoggingBuildStep):
 
         def cb(rclist):
             rc = SUCCESS # (this rc is not the same variable as that above)
-            for was_cb, buildsetstatus in rclist:
+            for was_cb, results in rclist:
                 # TODO: make this algo more configurable
                 if not was_cb:
                     rc = EXCEPTION
+                    log.err(results)
                     break
-                if buildsetstatus.getResults() == FAILURE:
+                if results == FAILURE:
                     rc = FAILURE
             return self.finished(rc)
 
